unit CPUInfo;

interface
uses Types, Windows, Dialogs, SysUtils;

function CPUClock(const intDelayTime: integer = 500): double;
function GetCPUFreq: Double;
function GetCPUID(Func1, Func2: DWORD): DWORD;
function GetCPUVendorID: string;
function GetProcessorCount: Integer;
function IsEnableCPUID: Boolean;
function IsHyperThreading: Integer;

implementation

function IsEnableCPUID: Boolean;
var
  dwTemp1, dwTemp2: DWORD;
begin
  asm
    pushfd
    pop eax
    mov dwTemp1,eax
    xor eax,00200000h
    push eax
    popfd
    pushfd
    pop eax
    mov dwTemp2,eax
end;
  result := (dwTemp1 <> dwTemp2);
end;

function GetCPUFreq: Double;
var
  dwEAX1, dwEAX2, dwEDX1, dwEDX2, dwTime:DWORD;
  {IsEnableRDTSC Start}
  // IsProcessorFeaturePresent()
  // NTnOSłgȂ
  function IsEnableRDTSC: Boolean;
  var
    dwTemp:DWORD;
  begin
    result := False;
    if not IsEnableCPUID then
      Exit;
    asm
      pushad
      mov eax, 1
      DB 0Fh, 0A2h // cpuid
      mov dwTemp, edx
      popad
    end;
    result := ((dwTemp and $00000010) = $00000010);
  end;
  {IsEnableRDTSC End}
begin
  result := -1;
  if not IsEnableRDTSC then
    Exit;
  dwTime := GetTickCount;
  asm
    DB 0Fh,031h // rdtsc
    mov dwEAX1,eax
    mov dwEDX1,edx
  end;
  while True do // 2bAChp̖[v
  begin
    if(GetTickCount - dwTime) > 2000 then
      Break;
  end;
  dwTime := GetTickCount - dwTime;
  asm
    DB 0Fh,031h //  rdtsc
    mov dwEAX2,eax
    mov dwEDX2,edx
  end;
  dwEDX2 := dwEDX2 - dwEDX1;
  dwEAX2 := dwEAX2 - dwEAX1;
  result := (dwEDX2 shl 32) + dwEAX2; //Clock
  result := result / (dwTime * 1000); //MHz
end;

function GetProcessorCount: Integer;
var
  SI:TSystemInfo;
begin
  GetSystemInfo(SI);
  result := SI.dwNumberOfProcessors;
end;

function GetCPUID(Func1, Func2: DWORD): DWORD;
var
  dwEAX, dwEBX, dwECX, dwEDX: DWORD;
begin
  result := $FFFFFFFF;
  if not IsEnableCPUID then
    Exit;
  asm
    pushad
    mov eax, Func1
    DB 0Fh, 0A2h // cpuid
    mov dwEAX, eax
    mov dwEBX, ebx
    mov dwECX, ecx
    mov dwEDX, edx
    popad
  end;
  case Func2 of
    0: result := dwEAX;
    1: result := dwEBX;
    2: result := dwECX;
    3: result := dwEDX;
  end;
end;

function IsHyperThreading: Integer;
var
  dw: DWORD;
begin
  dw := GetCPUID(1, 3);
  dw := dw shr 28;
  dw := dw and $00000001;
  if dw = 0 then
  begin
    Result := -1;
  end else begin
    dw := GetCPUID(1, 1);
    dw := dw shr 16;
    dw := dw and $000000FF;
    Result := dw;
  end;
end;

function GetCPUVendorID: string;
var
  dw1, dw2: DWORD;
  s: string;

  procedure Get4Characters;
  var
    i: Integer;
  begin
    for i := 0 to 3 do
    begin
      dw2 := dw1 and $000000FF;
      s := s + Char(dw2);
      dw1 := dw1 shr 8;
    end;
  end;
begin
  s := '';

  dw1 := GetCPUID(0, 1);
  Get4Characters;
  dw1 := GetCPUID(0, 3);
  Get4Characters;
  dw1 := GetCPUID(0, 2);
  Get4Characters;

  Result := s;
end;

// returns CPU clock frequency
// intDelayTime: measure time in ms - increase to get a more accuracy return
function CPUClock(const intDelayTime: integer = 500): double;
var
  TimerHi, TimerLo: DWORD;
  PriorityClass, Priority: Integer;
begin
  // saves thread priority for the process
  PriorityClass := GetPriorityClass(GetCurrentProcess);
  Priority := GetThreadPriority(GetCurrentThread);

  // sets priority to REAL-TIME
  SetPriorityClass(GetCurrentProcess, REALTIME_PRIORITY_CLASS);
  SetThreadPriority(GetCurrentThread, THREAD_PRIORITY_TIME_CRITICAL);

  // "delay" for priority effect
  Sleep(10);

  asm
    dw 310Fh // rdtsc
    mov TimerLo, eax
    mov TimerHi, edx
  end;

  // waits for calculations
  Sleep(intDelayTime);

  asm
    dw 310Fh // rdtsc
    sub eax, TimerLo
    sbb edx, TimerHi
    mov TimerLo, eax
    mov TimerHi, edx
  end;

  // restores process priority
  SetThreadPriority(GetCurrentThread, Priority);
  SetPriorityClass(GetCurrentProcess, PriorityClass);

  // sets the result with CPU clock frequency
  result := TimerLo / (1000.0 * intDelayTime);
end;
end.
