unit frmMain;
interface
uses
  Windows, Messages, SysUtils, Variants, Classes,
  Graphics, Controls, Forms, Dialogs, Menus,
  ShellAPI, IniFiles, ShlObj, StdCtrls, ExtCtrls,
  UnitMSec, PrimeSearch, frmGather;

const
  THREAD_MAX = 63;
type
  TFormMain = class(TForm)
    MenuMain: TMainMenu;
    MenuFile: TMenuItem;
    MenuHelp: TMenuItem;
    MenuFExit: TMenuItem;
    MenuHelpWeb: TMenuItem;
    MenuHelpReadme: TMenuItem;
    MenuHelpVersion: TMenuItem;
    MenuHelpStartMenu: TMenuItem;
    MenuHelpSep1: TMenuItem;
    PanelTop: TPanel;
    MemoResult: TMemo;
    LabelMax: TLabel;
    ComboBoxEndValue: TComboBox;
    LabelThreadMax: TLabel;
    ComboBoxThreadMax: TComboBox;
    CheckBoxThreadIncrease: TCheckBox;
    CheckBoxSaveToFile: TCheckBox;
    ButtonStart: TButton;
    SaveDialog1: TSaveDialog;
    MenuFileSaveTestResult: TMenuItem;
    MenuFileResultGather: TMenuItem;
    procedure FormCreate(Sender: TObject);
    procedure MenuFExitClick(Sender: TObject);
    procedure MenuHelpVersionClick(Sender: TObject);
    procedure MenuHelpWebClick(Sender: TObject);
    procedure MenuHelpReadmeClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action:
      TCloseAction);
    procedure MenuHelpStartMenuClick(Sender: TObject);
    procedure ComboBoxThreadMaxChange(Sender: TObject);
    procedure ButtonSaveToFileClick(Sender: TObject);
    procedure ButtonStartClick(Sender: TObject);
    procedure MenuFileSaveTestResultClick(Sender: TObject);
    procedure MemoResultChange(Sender: TObject);
    procedure MenuFileResultGatherClick(Sender: TObject);
  private
    OutputString: string;
    TimeStart, TimeEnd, TimeTaken, TimeBase: Double;
    WaitIdle: Integer;
    StartValue, EndValue: Int64;
    PrimeSearchArray: array[0..THREAD_MAX] of TPrimeSearch;
    ThreadFinished: array[0..THREAD_MAX] of Boolean;

    procedure ControlEnable(Value: Boolean);
    procedure CallPrimeSearchThread(ThreadIndex: Integer;
      RangeLow, RangeHigh, BasePrimeHigh: Cardinal);

    procedure Output;
    procedure PrimeSearchThreadDone(Sender: TObject);
    procedure SaveResultToFile;
    procedure SearchPrime(RangeLow, RangeHigh, BasePrimeHigh: Cardinal);
    procedure StartCalc(LocalThreadMax: Integer);
    procedure StartThreads;
    procedure IdleProc(Sender: TObject; var Done: Boolean);
  public
    { Public 錾 }
    //PrimeList: TList;
    PrimeArray: array of Cardinal;
    PrimeArrayWriteBuffer: array[0..THREAD_MAX] of array of Cardinal;
    ThreadMax: Integer;

  end;
var
  FormMain: TFormMain;

implementation

{$R *.dfm}
uses hhUnitWindowsVersion, hhUnit01, frmSMenu, Math, CPUInfo;

const
  VERSION = 0.12;
  LAST_UPDATE = '2004/07/31';

  FORM_POSITION_DIFFERENCE = 40;
  PROGRAM_NAME = 'MThread2';
  README_FILENAME = 'README.TXT';
  SHORTCUT_NAME = PROGRAM_NAME + '.lnk';
  SHORTCUT_COMMENT = PROGRAM_NAME;
  URL = 'http://arena.nikkeibp.co.jp/winpc/';

  HORIZONTAL_LINE = '-----------------------------------------------------------';

  INITIAL_PRIMES: array[0..167] of Cardinal =
    (2, 3, 5, 7, 11, 13, 17, 19, 23, 29,
     31, 37, 41, 43, 47, 53, 59, 61, 67, 71,
     73, 79, 83, 89, 97, 101, 103, 107, 109, 113,
     127, 131, 137, 139, 149, 151, 157, 163, 167, 173,
     179, 181, 191, 193, 197, 199, 211, 223, 227, 229,
     233, 239, 241, 251, 257, 263, 269, 271, 277, 281,
     283, 293, 307, 311, 313, 317, 331, 337, 347, 349,
     353, 359, 367, 373, 379, 383, 389, 397, 401, 409,
     419, 421, 431, 433, 439, 443, 449, 457, 461, 463,
     467, 479, 487, 491, 499, 503, 509, 521, 523, 541,
     547, 557, 563, 569, 571, 577, 587, 593, 599, 601,
     607, 613, 617, 619, 631, 641, 643, 647, 653, 659,
     661, 673, 677, 683, 691, 701, 709, 719, 727, 733,
     739, 743, 751, 757, 761, 769, 773, 787, 797, 809,
     811, 821, 823, 827, 829, 839, 853, 857, 859, 863,
     877, 881, 883, 887, 907, 911, 919, 929, 937, 941,
     947, 953, 967, 971, 977, 983, 991, 997);
  START_VALUE = 1000;


procedure TFormMain.MenuHelpVersionClick(Sender: TObject);
var
  hhWindowsVersion: ThhWindowsVersion;
begin
  hhWindowsVersion := ThhWindowsVersion.Create(Self);
  try
    ShowMessage(PROGRAM_NAME + #13#10 +
      Format('Version %1.2f (', [VERSION]) +
      LAST_UPDATE + ')' + #13#10 +
      '(c)2004 Hideo Harada All rights reserved.'
      + #13#10 + #13#10 +
      'OS: ' + hhWindowsVersion.Text + #13#10 +
      'User/Computer: ' + hhGetUserName + '/' +
      hhGetComputerName);
  finally
    hhWindowsVersion.Free;
  end;
end;

procedure TFormMain.FormCreate(Sender: TObject);
var
  i: Integer;
  s: string;
  Ini: TIniFile;
begin
  Self.Caption := PROGRAM_NAME;
  PanelTop.Align := alTop;
  PanelTop.Caption := '';
  MemoResult.Align := alClient;
  MemoResult.Text := '';
  MemoResultChange(Self);
  MemoResult.Font.Size := 12;
  MemoResult.ScrollBars := ssVertical;

  SaveDialog1.Filter := 'eLXgt@C(*.txt)|*.txt|ׂẴt@C(*.*)|*.*';
  SaveDialog1.InitialDir := ExtractFilePath(Application.ExeName);
  SaveDialog1.DefaultExt := 'txt';

  StartValue := START_VALUE;

  for i := 1 to 22 do
    ComboBoxEndValue.Items.Add(FloatToStr(IntPower(2, i) * StartValue));
  ComboBoxEndValue.ItemIndex := 0;

  for i := 0 to THREAD_MAX + 1 do
    ComboBoxThreadMax.Items.Add(IntToStr(i));

  //INIt@C̏
  s := (ChangeFileExt(Application.ExeName, '.ini'));
  Ini := TIniFile.Create(s);
  with Ini do
  begin
    Self.Top := ReadInteger('Window', 'Top', 40);
    Self.Left := ReadInteger('Window', 'Left', 100);
    Self.Width := ReadInteger('Window', 'Width', 600);
    Self.Height := ReadInteger('Window', 'Height', 500);
    ComboBoxEndValue.ItemIndex := ReadInteger('Parameter', 'EndValue', 0);
    if ReadBool('Parameter', 'ThreadIncrease', False) then
      CheckBoxThreadIncrease.Checked := True;
    if ReadBool('Parameter', 'SaveToFile', False) then
      CheckBoxSaveToFile.Checked := True;
    ComboBoxThreadMax.ItemIndex := ReadInteger('Parameter', 'ThreadMax', 0);
    ComboBoxThreadMaxChange(Self);
    Free;
  end;


  OutputString := '';

  //ACh
  Application.OnIdle := IdleProc;
  WaitIdle := -1; //-1ŃXgbvC̐Ń[hw
end;

procedure TFormMain.MenuFExitClick(Sender: TObject);
begin
  Self.Close;
end;

procedure TFormMain.MenuHelpWebClick(Sender: TObject);
var
  Return: Integer;
begin
  Return := ShellExecute(Application.Handle, 'open',
    PChar(URL), '', '', SW_SHOWNORMAL);
  if Return <= 32  then
    ShowMessage(URL + 'J܂ł');
end;

procedure TFormMain.MenuHelpReadmeClick(Sender: TObject);
var
  FullPath: String;
  Return: Integer;

begin
  FullPath := ExtractFilePath(Application.ExeName) +
    README_FILENAME;
  Return := ShellExecute(Application.Handle, 'open',
    PChar(FullPath), '', '', SW_SHOWNORMAL);
  if Return <= 32  then
    ShowMessage(FullPath + 'J܂łB' + #13#10 +
      'Windowst̃inotepad.exejȂǂŊJĂ݂Ă');
end;

procedure TFormMain.FormClose(Sender: TObject; var Action: TCloseAction);
var
  Ini: TIniFile;
  s: string;
begin
  s := (ChangeFileExt(Application.ExeName, '.ini'));
  try
    Ini := TIniFile.Create(s);
    With Ini do
    begin
      WriteInteger('Window', 'Top', Self.Top);
      WriteInteger('Window', 'Left', Self.Left);
      WriteInteger('Window', 'Width', Self.Width);
      WriteInteger('Window', 'Height', Self.Height);
      WriteInteger('Parameter', 'EndValue', ComboBoxEndValue.ItemIndex);
      WriteInteger('Parameter', 'ThreadMax', ComboBoxThreadMax.ItemIndex);
      WriteBool('Parameter', 'ThreadIncrease', CheckBoxThreadIncrease.Checked);
      WriteBool('Parameter', 'SaveToFile', CheckBoxSaveToFile.Checked);

      UpdateFile; //Windowsł͂ȂĂ
      Free;
    end;
  except
    //Ini.Free;
  end;
end;

procedure TFormMain.MenuHelpStartMenuClick(Sender: TObject);
var
  hhWindowsVersion: ThhWindowsVersion;
  IsNT, IsAllUser: Boolean;
  s: string;
  r: LongBool;
begin
  //NTǂ̏Ԃl
  hhWindowsVersion := ThhWindowsVersion.Create(Self);
  try
    IsNT := hhWindowsVersion.IsNT;
  finally
    hhWindowsVersion.Free;
  end;

  if IsNT then //NTȂI_CAO\
  begin
    with FormStartMenu do
    begin
      Left := Self.Left + FORM_POSITION_DIFFERENCE;
      Top := Self.Top + FORM_POSITION_DIFFERENCE;
      ShowModal;
    end;
    if FormStartMenu.IsCancel then
      Exit
    else
      IsAllUser := FormStartMenu.IsAllUser;
  end
  else
    //98łCSIDL_PROGRAMSłȂ΂ȂȂBC
    //łʂƂĂ͑S[U[L̃j[ڂɂȂ
    IsAllUser := False;

  SetLength(s, MAX_PATH);
  if IsAllUser then
    r := SHGetSpecialFolderPath(Handle, PChar(s),
      CSIDL_COMMON_PROGRAMS, False)
  else
    r := SHGetSpecialFolderPath(Handle, PChar(s),
      CSIDL_PROGRAMS, False);

  if r = True then
  begin
    try
      hhMakeShortcut(PChar(s) + '\' + SHORTCUT_NAME,
        Application.ExeName, SHORTCUT_COMMENT);
    except
      ShowMessage('G[FV[gJbg쐬ł܂ł');
    end;
  end
  else
    ShowMessage('G[FSHGetSpecialFolderPath̎sɎs܂');
end;

procedure TFormMain.IdleProc(Sender: TObject; var Done: Boolean);
begin
  case WaitIdle of
    -1:;
    0..THREAD_MAX:
      begin
        Application.ProcessMessages;
        Caption := PROGRAM_NAME + '(ҋ@)';
        Sleep(3000);
        Caption := PROGRAM_NAME + '(' + IntToStr(WaitIdle + 1) + ')';
        Self.Repaint;
        Application.ProcessMessages;
        
        TimeStart := now_ms;
        StartCalc(WaitIdle);
      end;
  end;
end;

procedure TFormMain.ComboBoxThreadMaxChange(Sender: TObject);
begin
  if ComboBoxThreadMax.ItemIndex < 1 then
  begin
    CheckBoxThreadIncrease.Checked := False;
    CheckBoxThreadIncrease.Enabled := False;
  end else begin
    CheckBoxThreadIncrease.Enabled := True;
  end;
end;

procedure TFormMain.ButtonSaveToFileClick(Sender: TObject);
begin
  SaveDialog1.Filter := 'eLXgt@C(*.txt)|*.txt|ׂẴt@C(*.*)|*.*';
  SaveDialog1.InitialDir := ExtractFilePath(Application.ExeName);

  if SaveDialog1.Execute then
  begin
    MemoResult.Lines.SaveToFile(SaveDialog1.FileName);
  end;
end;

procedure TFormMain.CallPrimeSearchThread(ThreadIndex: Integer; RangeLow,
  RangeHigh, BasePrimeHigh: Cardinal);
begin
  PrimeSearchArray[ThreadIndex] := TPrimeSearch.Create(True);
  PrimeSearchArray[ThreadIndex].Index := ThreadIndex;
  PrimeSearchArray[ThreadIndex].OnTerminate := PrimeSearchThreadDone;
  PrimeSearchArray[ThreadIndex].FreeOnTerminate := True;
  PrimeSearchArray[ThreadIndex].RangeLow := RangeLow;
  PrimeSearchArray[ThreadIndex].RangeHigh := RangeHigh;
  PrimeSearchArray[ThreadIndex].BasePrimeHigh := BasePrimeHigh;
  PrimeSearchArray[ThreadIndex].Resume;
end;

procedure TFormMain.Output;
begin
  TimeEnd := now_ms;

  if CheckBoxSaveToFile.Checked then
  begin
    SaveResultToFile;
  end;

  TimeTaken := TimeEnd - TimeStart;
  if ThreadMax = -1 then TimeBase := TimeTaken;

  if TimeBase = 0 then
    OutputString := OutputString + IntToStr(ThreadMax + 1) + #09 +
      IntToStr(High(PrimeArray) + 1) + #09 +
      TimeConvert(1, TimeTaken) + #09 +
      '-' + #13#10
  else
    OutputString := OutputString + IntToStr(ThreadMax + 1) + #09 +
      IntToStr(High(PrimeArray) + 1) + #09 +
      TimeConvert(1, TimeTaken) + #09 +
      Format('%f', [100 / (TimeTaken / TimeBase)]) + #13#10;

  SetLength(PrimeArray, 0);

  if CheckBoxThreadIncrease.Checked then
  begin
    Inc(ThreadMax);
    if ThreadMax < ComboBoxThreadMax.ItemIndex then
    begin
      WaitIdle := ThreadMax;
    end else begin
      WaitIdle := -1;
      OutputString := OutputString +
        HORIZONTAL_LINE + #13#10;
      ControlEnable(True);
    end
  end else begin
    OutputString := OutputString +
      HORIZONTAL_LINE + #13#10;
    ControlEnable(True);
  end;
  MemoResult.Lines.Text := OutputString;
end;

procedure TFormMain.PrimeSearchThreadDone(Sender: TObject);
var
  i, j, NewPrimeCount: Integer;
  Index: Integer;
begin
  ThreadFinished[TPrimeSearch(Sender).Index] := True;
  for i := 0 to ThreadMax do
    if Not ThreadFinished[i] then Exit;

  if ThreadMax > 0 then
  begin
    NewPrimeCount := 0;
    for i := 0 to ThreadMax do
      NewPrimeCount := NewPrimeCount + Length(PrimeArrayWriteBuffer[i]);

    Index := High(PrimeArray) + 1;
    SetLength(PrimeArray, Length(PrimeArray) + NewPrimeCount);

    for i := 0 to ThreadMax do
      for j := 0 to High(PrimeArrayWriteBuffer[i]) do
      begin
        PrimeArray[Index] := PrimeArrayWriteBuffer[i][j];
        Inc(Index);
      end;
  end;

  StartValue := StartValue * 2;
  if StartValue >= EndValue then
    Output
  else
    StartThreads;
end;

procedure TFormMain.SaveResultToFile;
var
  c: Cardinal;
  FileName: string;
  f: TextFile;
begin
  FileName := ExtractFilePath(Application.ExeName) +
    'Prime' + Format('%.2d', [ThreadMax + 1]) + '.txt';

  DeleteFile(FILENAME);
  AssignFile(f, FILENAME);
  Rewrite(f);
  Append(f);

  for c := 0 to High(PrimeArray) do
    Write(f, IntToStr(PrimeArray[c]) + #13#10);
  CloseFile(f);
end;

procedure TFormMain.SearchPrime(RangeLow, RangeHigh,
  BasePrimeHigh: Cardinal);
var
  MyBits: TBits;
  c, d: Cardinal;
  i, Count, OldHigh: Integer;
begin
  MyBits := TBits.Create;
  try
    MyBits.Size := RangeHigh - RangeLow + 1;

    for c := 0 to BasePrimeHigh do
    begin
      if PrimeArray[c] > Trunc(Sqrt(RangeHigh)) then Break;
      for d := (RangeLow div PrimeArray[c]) to
        (RangeHigh div PrimeArray[c]) do
      begin
        i := PrimeArray[c] * d - RangeLow;
        if i >= 0 then MyBits[i] := True;
      end;
    end;

    Count := 0;
    OldHigh := High(PrimeArray);
    for c := 0 to MyBits.Size - 1 do
      If Not MyBits[c] then Inc(Count);
    SetLength(PrimeArray, Length(PrimeArray) + Count);

    Count := 1;
    for c := 0 to MyBits.Size - 1 do
      if Not MyBits[c] then
      begin
        PrimeArray[OldHigh + Count] := RangeLow + c;
        Inc(Count);
      end;
  finally
    MyBits.Free;
  end;
end;

procedure TFormMain.StartThreads;
var
  i: Integer;
  rl, rh: Integer;
begin
  for i := 0 to THREAD_MAX do
  begin
    SetLength(PrimeArrayWriteBuffer[0], 0);
    ThreadFinished[i] := False;
  end;

  for i := 0 to ThreadMax do
  begin
    rl := StartValue * (ThreadMax + i + 1) div (ThreadMax + 1) + 1;
    rh := StartValue * (ThreadMax + i + 2) div (ThreadMax + 1);
    CallPrimeSearchThread(i, rl, rh, High(PrimeArray));
  end;
end;

procedure TFormMain.ButtonStartClick(Sender: TObject);
var
  hhWindowsVersion: ThhWindowsVersion;
  s, m1, m2: string;
  MemoryStatus:TMemoryStatus;
  Count, Size: Integer;
  TabWidth: longint;
begin
  TabWidth := 16 * LoWord(GetDialogBaseUnits) div 2;
  MemoResult.Perform(EM_SETTABSTOPS, 1, longint(@TabWidth));
  MemoResult.Invalidate;

  ControlEnable(False);
  TimeBase := 0;
  if IsHyperThreading = -1 then
    s := 'Ȃ'
  else
    s := IntToStr(IsHyperThreading);

  Count:= AllocMemCount;
  Size:= AllocMemSize;
  MemoryStatus.dwLength:= SizeOf(MemoryStatus);
  GlobalMemoryStatus(MemoryStatus);
  with MemoryStatus do
  begin
    m1 := Format('%.2fMB', [dwTotalPhys / (1024*1024)]);
    m2 := Format('%.2fMB', [dwAvailPhys / (1024*1024)]);
  end;

  hhWindowsVersion := ThhWindowsVersion.Create(Self);

  OutputString := 'Rs[^^[U[F' + hhGetComputerName +
    '^' + hhGetUserName + #13#10 +
    'CPUx_[IDF' + GetCPUVendorID + #13#10 +
    '_vZTF' + IntToStr(GetProcessorCount) + '@@' +
    'Hyper-ThreadingF' + s + '@@' +
    'gF' + Format('%.2f', [CPUClock]) + 'MHz' + #13#10 +
    '[F' + m1 + '@@' + '󂫕[F' + m2 + #13#10 +
    'OSF' + hhWindowsVersion.Text + #13#10 + #13#10 +
    'eXgJnF' + DateTimeToStr(Now) + #13#10 +
    'TlF' + ComboBoxEndValue.Items[ComboBoxEndValue.ItemIndex] + #13#10 + #13#10 +
    'qXbh' + #09 + 'f' + #09 + '' +
    #09 + '' + #13#10 + HORIZONTAL_LINE + #13#10;

  hhWindowsVersion.Free;

  if CheckBoxThreadIncrease.Checked then
  begin
    TimeStart := now_ms;
    StartCalc(-1);
  end else begin
    TimeStart := now_ms;
    StartCalc(ComboBoxThreadMax.ItemIndex - 1);
  end;
end;

procedure TFormMain.StartCalc(LocalThreadMax: Integer);
var
  i: Integer;
begin
  WaitIdle := -1;
  StartValue := START_VALUE;
  EndValue := Trunc(IntPower(2, ComboBoxEndValue.ItemIndex + 1) * StartValue);
  ThreadMax := LocalThreadMax;

  SetLength(PrimeArray, High(INITIAL_PRIMES) + 1);
  for i := Low(INITIAL_PRIMES) to High(INITIAL_PRIMES) do
    PrimeArray[i] := INITIAL_PRIMES[i];

  if LocalThreadMax = -1 then
  begin
    repeat
      SearchPrime(StartValue + 1, StartValue * 2, High(PrimeArray));
      StartValue := StartValue * 2;
    until StartValue >= EndValue;
    Output;
  end else begin
    StartThreads;
  end;
end;

procedure TFormMain.MenuFileSaveTestResultClick(Sender: TObject);
var
  s: string;
begin
  if MemoResult.Lines.Text = '' then Exit;

  s := ComboBoxEndValue.Items[ComboBoxEndValue.ItemIndex];
  s := Copy(s, 0, Length(s) - 3) + 'k.txt';
  s := SaveDialog1.InitialDir + '\' + s;

  SaveDialog1.FileName := s;
  if SaveDialog1.Execute then
  begin
    MemoResult.Lines.SaveToFile(SaveDialog1.FileName);
    SaveDialog1.InitialDir := ExtractFilePath(SaveDialog1.FileName);
  end;
end;

procedure TFormMain.ControlEnable(Value: Boolean);
begin
  ButtonStart.Enabled := Value;
  ComboBoxEndValue.Enabled := Value;
  ComboBoxThreadMax.Enabled := Value;
  CheckBoxThreadIncrease.Enabled := Value;
  CheckBoxSaveToFile.Enabled := Value;
  MenuFile.Enabled := Value;
  MenuHelp.Enabled := Value;
  ComboBoxThreadMaxChange(Self);
  if Value then Self.Caption := PROGRAM_NAME;
end;

procedure TFormMain.MemoResultChange(Sender: TObject);
begin
  if MemoResult.Lines.Text = '' then
    MenuFileSaveTestResult.Enabled := False
  else
    MenuFileSaveTestResult.Enabled := True;
end;

procedure TFormMain.MenuFileResultGatherClick(Sender: TObject);
begin
  FormGather.ShowModal;
end;

end.
